# 
# @author jonathan_lee@berkeley.edu (Jonathan N. Lee)
# 
# layer defines a interface layer and related functions such
# as forward and back propagation.
# 
# parameters are typically:
#   size - integer of size of layer (number of activations)
#   act - activation function as defined by activations module
# so the initializer is actually an implementation for declaration of
# instance variables
# weights and biases not set by default
#
# all layers implement the LayerIterator class for forward
# propagation iteration
# 
# if you add a new class the implements the layer interface, be sure to add
# add it and a unique string identifier to the list of types
#
# this class interface is not designed to be implemented.
#


# system libraries
# internal libraries
from neuralpy2.iterators import LayerIterator
# third party libraries
import numpy as np


class Layer(object):

    
    def __init__(self, size, activ):
        self.size = size        # num nodes in layer
        self.activ = activ      # activation function
        
        self.w = None           # incoming weights  (size x prev-size) matrix
        self.b = None           # incoming biases   (size x 1) vector
        
        self.a = None           # incoming activation vector from previous layer
        self.z = None           # weighted sum + bias vector
        self.x = None           # activation function on z vector

        self.next_ = None       # layer that comes after this
        self.prev = None        # layer that comes before this
        
        self.delta_w = None     # partial weights set to none as no update required
        self.delta_b = None     # partial biases set to none for same reason

        self.type_ = type_gen   # default generic layer (not be used)
    

    # propagate forward while recording vectors for optimization
    # given activations of previous layer
    def forward(self, incoming_activations):    
        raise NotImplementedError

    # propagate backward while using vectors from forward
    # propagation given mu vector to be passed to next layer
    def _backward(self, mu):
        raise NotImplementedError
    
    # use only this method to add a layer to the list
    # append function should be the default interface
    def append(self, next_):
        raise NotImplementedError

    # Do not call this method from a non-layer caller.
    # prepend function that should also establish
    # the weight connection between the adjacent layers.
    def _prepend(self, prev):
        raise NotImplementedError

    # handle randomization of weights and biases
    # based on the layer's situation (ie preceding layer)
    # and type of layer
    def randomize_parameters(self):
        raise NotImplementedError

    # update the weights and biases for the layer
    # using the calculated cost gradient components
    # this function may not be used by all layers.
    # the delta matrices should be the same shape
    # as the weight and bias matrices.
    # after update, deltas should be zeroed.
    def _apply_updates(self, alpha):
        if self.b is not None and self.w is not None:
            self.b -= alpha * self.delta_b
            self.w -= alpha * self.delta_w

    # zero the delta matrices by setting them equal to
    # matrices of the same shape filled with zeros.
    # if they are none, just leave them as None
    def _zero_deltas(self):
        if self.w is not None and self.b is not None:
            self.delta_w = np.zeros(self.w.shape)
            self.delta_b = np.zeros(self.b.shape)
    
    # iterable to iterate forwards on across layers
    # returns specific LayerIterator class
    def __iter__(self):
        return LayerIterator(self)

    # returns a random n-dimensional array of 
    # of random values using numpy where n
    # is the number of dimensions. Each argument
    # is the size of the respective dimension
    def rand(self, *args):
        return np.random.randn(*args)

    # iterate overself adding one to a counter
    # during each iteration to find the count
    # from this layer to the last layer. return count
    def __len__(self):
        it = iter(self)
        length = 0
        for next_ in it:
            length += 1
        return length

    # convert self to a format suitable for json
    # interpretation. Requires structural characteristics
    # of layer and parameters
    def toJSON(self):
        w, b = self.w, self.b
        if self.w is not None: w = self.w.tolist()
        if self.b is not None: b = self.b.tolist()
        data = {
            "size": self.size,
            "type": self.type_,
            "activ": str(self.activ),
            "w": w,
            "b": b
        }
        return data

# types of layers to specify layer type property
# available in layers module
type_gen = "gen"
type_input = "input"
type_mlp = "mlp"






# system libraries
# internal libraries
# third party libraries
import numpy as np

# Cost interface that abstracts both the normal,
# generic function and derivative of that function
# deriv is assumed to ALWAYS be the analytical derivative
class Cost():
    def func(self, *args):
        raise NotImplementedError
    def deriv(self, *args):
        raise NotImplementedError

# implementation of cost interface for mean
# squared error which is defined as the one half the square
# of the difference between the actual and expected
class MeanSquare(Cost):
    
    # normal function defined as one
    # half of the mean square error
    def func(self, x, y):
        return 0.5 * ((x - y) ** 2)
    
    # derivative of the normal function
    def deriv(self, x, y):
        return x - y

# EXPORT
mean_square = MeanSquare()

mapping = {
    "mean_square": mean_square
}


from neural2 import *
from activations import *
import colors
from layers import *
import iterators

__author__ =    "Jonathan N Lee"
__license__ =   "MIT"
__version__ =   "1.2.0"
__email__ =     "jonathan_lee@berkeley.edu"


#
# @author jonathan_lee@berkeley.edu (Jonathan N. Lee)
# 
# module that implementes the iterator interface designed
# specifically to iterate over layers which are similar to
# linked lists.
#
# for layer to implement this class add:
# 
#   def __iter__(self):
#       return LayerIterator(<iterable>)
#


# system libraries
# internal libraries
from iterator import Iterator
# third party libraries

class LayerIterator(Iterator):

    # initializer a layer iterator just like
    # a linked list where the iterator has a root
    # object
    def __init__(self, root):
        self.root = root

    def __iter__(self):
        return self

    # return the next forward propagation of the layer
    # unless the layer is none then stop iteration.
    def next(self):
        if self.root is None:
            raise StopIteration
        else:
            temp = self.root
            self.root = self.root.next_
            return temp



# 
# @author jonathan_lee@berkeley.edu (Jonathan N. Lee)
# 
# Implementation of layer interface for
# fully connected feedforward multilayer perceptron
# retains super initializer
# weights not initialized by default but biases are as random
# append should be only interface for joining adjacent layers
# do NOT use prepend, append will prepend for you.
# prepend assigns random and appropriately-sized weights
# 

# system libraries
# internal libraries
import layer
# third party libraries
import numpy as np


class MLP(layer.Layer):

    # initialize mlp with type (as with all layers)
    # and initialize the random biases
    def __init__(self, size, activ):
        super(MLP, self).__init__(size, activ)
        self.type_ = layer.type_mlp
        self.b = self.rand(size, 1)
        self.delta_b = np.zeros(self.b.shape)
 

    # Implementation of layer interface for
    # fully connected feedforward multilayer perceptron
    def forward(self, incoming_activations):
        self.a = incoming_activations
        self.z = np.dot(self.w, self.a) + self.b
        self.x = self.activ.func(self.z)
        return self.x


    # Implementation of backward propagation
    # intended to compute the delta weights/biases of
    # incoming parameters for this layer only. Compute
    # the mu for the previous layer as well assuming there is
    # previous input to do so
    def _backward(self, mu):
        self.delta_b += mu
        self.delta_w += np.dot(mu, self.a.transpose())
        if self.prev.z is not None:
            mu = np.dot(self.w.transpose(), mu) * self.activ.deriv(self.prev.z)
            return mu
        else:
            return None


    # set next_ layer to instance var.
    # prepend self to next_.
    # caller beware: will potentially overwrite next_'s weights
    def append(self, next_):
        self.next_ = next_
        if next_ is not None:
            next_._prepend(self)


    # set prev layer to instance var.
    # initialize the weights as random between the layers
    # caller beware: will overwrite self's weights
    def _prepend(self, prev):
        self.prev = prev
        self.w = np.random.randn(self.size, prev.size)
        self._zero_deltas()


    # randomize parameters for fully connected
    # which is just random weights for each connectin
    # with preceding layer and full vector of random biases
    def randomize_parameters(self):
        if self.prev is not None:
            self.w = self.rand(self.size, prev.size)
        self.b = self.rand(self.size, 1)





# system libraries
# internal libraries
# third party libraries
import numpy as np


# Interface for activation function. Should implement
# func which is the regular function, deriv which is the
# derivative of the function, and optional fast_deriv which
# is an optimized form of div that takes a different param
class Activ(object):
    def func(self,x):
        raise NotImplementedError
    def deriv(self,x):
        raise NotImplementedError
    def fast_deriv(self,y):
        raise NotImplementedError
    def __repr__(self):
        raise NotImplementedError


# Implements activ for LINEAR function
# no fast derivative
class Lin(Activ):
    def func(self, x):
        return x
    
    def deriv(self, x):
        return 1.0

    def __repr__(self):
        return "linear"

    
# Implements activ for SIGMOID (logistic) function
# includes fast derivative
class Sig():
    def func(self, x):
        return 1.0 / (1.0 + np.exp(-x))
    
    def deriv(self, x):
        y = self.func(x)
        return y * (1.0 - y)

    def fast_deriv(self, y):
        return y * (1.0 - y)

    def __repr__(self):
        return "sigmoid"

# EXPORTS of instances
linear = Lin()
sigmoid = Sig()

mapping = {
    str(sigmoid): sigmoid,
    str(linear): linear
}



# 
# @author jonathan_lee@berkeley.edu (Jonathan N. Lee)
# 
# Don't bother reading this. These are just color codes.
#

header = '\033[95m'
blue = '\033[94m'
green = '\033[92m'
yellow = '\033[93m'
red = '\033[91m'
end = '\033[0m'
bold = '\033[1m'
underline = '\033[4m'



# 
# @author jonathan_lee@berkeley.edu (Jonathan N. Lee)
#
# iterator interface to be implemented by iterator
# classes such as LayerIterator. Do no implement
# these functions.
# makes implemented classes iterable.
#

# system libraries
# internal libraries
# third party libraries


class Iterator(object):
    def __init__(self, *args):
        raise NotImplementedError
    def __iter__(self):
        raise NotImplementedError
    def next(self):
        raise NotImplementedError


#
# @author jonathan_lee@berkeley.edu (Jonathan N. Lee)
# 
# NetworkBase module that defines an abstract network. sizes parameter determines
# the number of nodes in each respective layer. layers by default
# consist of an inputs and mlps with sigmoid activations.
# 


# system libraries
import random
import json
# internal libraries
import layers
import activations
import colors
import costs
# third party libraries
import numpy as np
import matplotlib.pyplot as plt

def output(s=""):
    print colors.green + str(s) + colors.end

# list2vec converts a list of any dimension to a numpy array
# that is also a column vector (mx1 matrix where m equals the
# number of elements in li. also will put non-lists into a list
def list2vec(li):
    li = np.array(li)
    return np.reshape(li, (len(li), 1))


# reconstruct json from a given file
# in the form of the saved Network structure
# see "save" function
# returns a constructed network model
def load(path):
    with open(path) as in_file:
        layers = json.load(in_file)
    net = Network( [ layers[0]['size'] ] )
    for layer in layers[1:]:
        size = layer['size']
        activ = layer['activ']
        type_ = layer['type']
        net.append(size, type_, activ)
        if layer['b'] is not None:  net.end.b = np.array(layer['b'])
        else:                       net.end.b = None
        if layer['w'] is not None:  net.end.w = np.array(layer['w'])
        else:                       net.end.w = None
    return net
        


class NetworkBase(object):

    # initialize the network by creating
    # a layers with the corresponding number of nodes
    # @params sizes     list of positive integers representing
    #                   nodes per layer
    def __init__(self, sizes):
        raise NotImplementedError


    # propagate forward by iterating over all layers
    # staring with the first layer
    # @param x          input column vector
    def forward(self, x):
        raise NotImplementedError

    # create and append layer to the end layer of the network
    # simply by using the end instance variable
    # see layers.py and activations.py for mappings
    # @param size       number of nodes in the layer
    # @param type_      layer identifier type    
    # @param activ      activation identifier type
    # type_ and activ are optional
    def append(self, *args):
        raise NotImplementedError

    # pop the last layer off the network
    # by setting end instance var to the 
    # previous layer and removing the previous
    # layers forward propagation to this
    def pop(self):
        raise NotImplementedError

    # train neural network based on training data
    # to optimize the weights. Abstract method
    # intended to be handled by implementing network
    # class
    # @param training_set       training_set which is a list
    #                           of tuples: first component is column
    #                           input vec, second is column output vec
    # @param epochs             number of iterations to update weights
    # @param alpha              learning rate
    # @param mini_batch_size    by default, reverts to 1
    # @param monitor            by default, no loss surveillence
    def train(self, *args):
        raise NotImplementedError

    # iterate through layer and update the biases
    # and weights as each layer type should handle
    # the individual implementation uniquely
    # applying the delta updates should always zero
    # out the delta weights
    # @param alpha              learning rate
    def _apply_updates(self, alpha):
        raise NotImplementedError

    # iterate through each alyer and zero out
    # the weights, as handled by the implementations
    # of the given layers
    def _zero_deltas(self):
        raise NotImplementedError

    # iterate through each layer and save the weights
    # and biases to a specified text file in the file_path
    # will also store the network hyperparameters
    # @param file_path      path to the write file
    def save(self, file_path):
        raise NotImplementedError

    # load weights and biases from a specified file path
    # Note: this overwrites the current network structure
    # and weights and biases. Use carefully.
    # @param file_path      path to the read file
    def load(self, file_path):
        raise NotImplementedError

    # compute the cost given a 
    def _compute_cost(self, training_set):
        raise NotImplementedError

    # convert the network to json format, essentially
    # just by saving information in string/number format
    def toJSON(self):
        raise NotImplementedError

    # save the network by structuring the network
    # as a list of dictionaries. dictionaries would
    # then easily be converted to json and later
    # reconstructed into a Network object
    def save(self, path):
        raise NotImplementedError



# Network class provides the implementation of NetworkBase's 
# non implemented functions. See interface class for details about
# each funciton (no comments in this class provided)
class Network(NetworkBase):


    def __init__(self, sizes):
        it = iter(sizes)
        self.start = layers.Input(next(it))              
        layer = self.start
        for size in it:
            layer.append( layers.MLP(size, activations.sigmoid) )
            layer = layer.next_
        self.end = layer


    def forward(self, x):
        x = list2vec(x)
        it = iter(self.start)
        for layer in it:
            x = layer.forward(x)
        return x


    def append(self, *args):
        size = args[0]
        if len(args) > 1:   type_ = args[1]
        else:               type_ = "mlp"       # default argument
        if len(args) > 2:   activ = args[2]
        else:               activ = "sigmoid"   # default argument
        
        type_ = layers.mapping[type_]
        activ = activations.mapping[activ]
        layer = type_(size, activ)
        self.end.append(layer)
        self.end = self.end.next_


    def pop(self):
        self.end = self.end.prev
        self.end.append(None)


    def train(self, training_set, epochs, alpha, mini_batch_size=1, monitor=False):
        training_set = [ ( list2vec(x), list2vec(y) ) for x, y in training_set]
        self.costcurve = []
        alpha = float(alpha)
        n = len(training_set)
        self._zero_deltas()
        for j in xrange(epochs):
            random.shuffle(training_set)
            mini_batches = [
                training_set[k:k + mini_batch_size] 
                for k in xrange(0, n, mini_batch_size)
            ]
            for mini_batch in mini_batches:
                self._update_batch(mini_batch, alpha)
            if monitor:
                output("Epoch " + str(j) + " compeleted")
                self.costcurve.append(self._compute_cost(training_set))

    def show_cost(self):
        x = xrange(0, len(self.costcurve), 1)
        plt.plot(x, self.costcurve)
        plt.show()


    def _update_batch(self, training_set, alpha):
        for x, y in training_set:
            self.forward(x)
            mu = costs.mean_square.deriv(self.end.x, y) *\
                self.end.activ.deriv(self.end.z)
            root = self.end
            while root is not None:
                mu = root._backward(mu)
                root = root.prev
        self._apply_updates(alpha/len(training_set))


    def _apply_updates(self, alpha):
        it = iter(self.start)
        for layer in it:
            layer._apply_updates(alpha)
            layer._zero_deltas()


    def _zero_deltas(self):
        it = iter(self.start)
        for layer in it:
            layer._zero_deltas()


    def _compute_cost(self, training_set):
        cost = 0
        for x, y in training_set:
            actual = self.forward(x)
            cost += np.linalg.norm(costs.mean_square.func(actual, y))
        return cost / len(training_set)

    def __len__(self):
        return len(self.start)
        
    
    def toJSON(self):
        it = iter(self.start)
        layers = []
        for layer in it:
            layers.append(layer.toJSON())
        return layers

        
    def save(self, path):
        netJSON = self.toJSON()
        with open(path, 'w') as outfile:
            json.dump(netJSON, outfile)
    

    """# --TODO--
    def save(path):
        it = iter(self.start)
        layers = []
        for layer in it:
            layers.append(layer)
        with open(path, 'w') as outfile:
            json.dump(layers, outfile)
"""



# 
# @author jonathan_lee@berkeley.edu (Jonathan N. Lee)
# 
# Implementation of layer interface for
# input layer. retains super intitializer.
# no weight or bias caculations
# no backward propagation
# only append, no prepend (assuming this is the first layer)
# 

# system libraries
# internal libraries
import layer
# third party libraries
import numpy as np


class Input(layer.Layer):
    
    def __init__(self, size):
        super(Input, self).__init__(size, None)
        self.type_ = layer.type_input

    # Implementation of layer interface
    # for input vector layer, no calculations
    def forward(self, x):
        self.x = x
        return x

    # do nothing becaus ethere is no backward
    # propagation calculation for input layer
    # return None to indicate end of propagation
    def _backward(self, *args):
        return None

    # assign to instance var next_
    # prepend self to next_
    # caller beware: may potentially overwrite next_'s weights
    def append(self, next_):
        self.next_ = next_
        next_._prepend(self)


    # there are no incoming parameters for the 
    # input layer
    def randomize_parameters(self):
        self.w = None
        self.b = None